#ifndef PSRFITS_HH_
#define PSRFITS_HH_
#include <fitsio.h>
#include <string>
#include "pvo-block.hh"
#include "pvo-header.hh"

using namespace Block;
using namespace Header;

/**
 * The namespace to control all things related to reading and writing
 * fits files in the PSRFITS format.
 */ 

namespace FITS
{

  /** Has a function worked or not? */
  enum RESULT { SUCCESS, FAILURE };
	
  /** The state of the file */
  enum STATE { OPEN,  ///< file is open
	       CLOSED ///< file is closed
  };

  /** 
   * Report a CFITSIO error, along with the offending function.  
   * An error message is printed, along with the relevant CFITSIO error
   * code and message.  
   * \param function The function from which the error was called.  
   * \param status The status value as returned by the CFITSIO function. 
   *  Generally, if status != 0, an error has occurred.
   */
  inline void FITSError(std::string function, int status){
    //    if(FITS::verbose){
    std::cerr << "\aFITSError <" << function << "> (status="<<status<<"): ";
    fits_report_error(stderr, status);
  }


  //===========================================================

  /**
   *  A class to keep track of the header keywords for the SUBINT
   *  binary table extension.
   */
  class Subint
  {
  public:


    Subint(){
      allocated = false;
      nch_strt = 0;
      int_type = "TIME";
      int_unit = "SEC";
    };
    virtual ~Subint(){
      if(allocated){
	delete [] dat_freq;
	delete [] dat_wts;
	delete [] dat_offs;
	delete [] dat_scl;
      }
    };

    friend class PSRFits;

    long   getNch_file(){return nch_file;};
    void   setNch_file(long l){nch_file=l;};
    long   getNch_strt(){return nch_strt;};
    void   setNch_strt(long l){nch_strt=l;};
    long   getNbits(){return nbits;};
    void   setNbits(long l){nbits=l;};
    long   getNbin(){return nbin;};
    void   setNbin(long l){nbin=l;};
    long   getNpol(){return npol;};
    void   setNpol(long l){npol=l;};
    long   getNsblk(){return nsblk;};
    void   setNsblk(long l){nsblk=l;};
    float  getTsamp(){return tsamp;};
    void   setTsamp(float f){tsamp=f;};
    std::string getInt_type(){return int_type;};
    void   setInt_type(std::string s){int_type=s;};
    std::string getInt_unit(){return int_unit;};
    void   setInt_unit(std::string s){int_unit=s;};
    bool   getAllocatedFlag(){return allocated;};
    void   setAllocatedFlag(bool b){allocated=b;};

    float *getDAT_freq(){return dat_freq;};
    void   setDAT_freq(float *f){dat_freq=f;};
    float *getDAT_wts(){return dat_wts;};
    void   setDAT_wts(float *f){dat_wts=f;};
    float *getDAT_offs(){return dat_offs;};
    void   setDAT_offs(float *f){dat_offs=f;};
    float *getDAT_scl(){return dat_scl;};
    void   setDAT_scl(float *f){dat_scl=f;};

    /** Return the size of the data array */
    long   getDataSize(){return nbits*nch_file*npol*nsblk;};

    /** Return the size of the data array in bytes */
    long   getDataSizeBytes(){return nbits*nch_file*npol*nsblk / 8;};

    /** A function that will write to a file the full set of keywords. */
    void writeKeywords(fitsfile *fptr);
 
  private:
    long        nbin;       ///< The SUBINT table keyword NBIN
    long        nbits;      ///< The SUBINT table keyword NBITS
    long        nch_strt;   ///< The SUBINT table keyword NCH_STRT
    long        nch_file;   ///< The SUBINT table keyword NCH_FILE
    long        npol;	    ///< The SUBINT table keyword NPOL
    long        nsblk;	    ///< The SUBINT table keyword NSBLK
    std::string int_type;   ///< The SUBINT table keyword INT_TYPE
    std::string int_unit;   ///< The SUBINT table keyword INT_UNIT
    float       tsamp;	    ///< The SUBINT table keyword TSAMP

    bool        allocated;  ///< Have the arrays been allocated yet?
    float      *dat_freq;   ///< The SUBINT table array DAT_FREQ
    float      *dat_wts;    ///< The SUBINT table array DAT_WTS
    float      *dat_offs;   ///< The SUBINT table array DAT_OFFS
    float      *dat_scl;    ///< The SUBINT table array DAT_SCL


  };

  
  const int SUBINT_kw_nch_strt = 0; 
  ///< The nch_strt keyword for the SUBINT table.

  const std::string SUBINT_kw_int_type = "TIME";
  ///< The int_type keyword for the SUBINT table.

  const std::string SUBINT_kw_int_unit = "SEC"; 
  ///< The int_unit keyword for the SUBINT table.
  
  const std::string PHU_kw_coord_md = "J2000.0"; 
  ///< The coord_md keyword for the Primary header.

  const double PHU_stt_offs_default = 0.;
  ///< The default value for the STT_OFFS PHU keyword. */
  
  /**
   * A class to deal with individual fits files.
   * This keeps track of all information relevant for a particular PSRFITS file.
   * All the functions that write to particular extensions are in here, as 
   *  well as general I/O routines.
   */ 

  class PSRFits
  {
  public:
    PSRFits(); ///<
    PSRFits(std::string name);
    virtual ~PSRFits(){};

    // Inline functions
    /** Is the file newly created? */
    bool        isNewFile(){return newFile;};
    void        setNewFile(){newFile = true;};
    void        setNewFile(bool f){newFile = f;};

    /** Is the file open or closed? */
    STATE       getFilestate(){return filestate;};

    /** Set the file state. */
    void        setFilestate(STATE S){filestate=S;};

    /** Get the UT of the first observations in the file. */
    std::string getUT(){return UT_initial;};

    /** Save the UT of the first observations. */
    void        setUT(std::string s){UT_initial=s;};

    /** Get the date of the first observations in the file. */
    std::string getDate(){return DATE_initial;};

    /** Save the date of the first observations. */
    void        setDate(std::string s){DATE_initial=s;};

    /** Get the beam number used for the observations. */
    int         getBeam(){return beam_number;};

    /** Save the beam number used for the observations. */
    void        setBeam(int i){beam_number=i;};

    /** Get the filename of the fits file. */
    std::string getFITSname(){return FITSname;};

    /** Manually set the filename of the fits file. */
    void        setFITSname(std::string name){FITSname = name;};

    /** Get the number of the first block stored in the file. */
    int         getInitialBlock(){return blk_initial;};

    /** Save the number of the first block stored in the file. */
    void        setInitialBlock(int block){blk_initial = block;};

    /** Opens file -- uses openFile(std::string) */
    int         openFile(){ return this->openFile(this->FITSname); };


    // Functions found in PSRFits.cc (in the same order...)
    /** Generate the name of the fits file. */
    void        setFITSname(Header::SURVEY surveyType); 

    /** Does the file exist on disk already? */
    bool        fileExists();                           
       
    /** Opens the file named by the argument. */
    int         openFile(std::string name);              
       
    /** Makes corrections to PSRFITS definition. */
    void        fixPSRFITS();  
       
    /** Writes new keywords to the primary fits header. */
    void        addNewKeywords(DataBlock &block); 
    
    /** Check whether the FITS file should be closed. */
    bool        fitsFull(DataBlock &block);
    long        getNumSamples(){return numSamples;};

    int         closeFile();   ///< Closes the file.
    int         gotoPHU();     ///< Moves to the primary header unit
    int         gotoHistory(); ///< Moves to the HISTORY binary table
			       ///   extension.
    int         gotoSubint();  ///< Moves to the SUBINT binary table
			       ///   extension.

    /** Delete a named binary table. */
    int         deleteTable(char tableName[]);

    /** Update the value of a keyword. */
    template <class T> int updateKeyword(std::string name, T value); 

    /** Return the value of a keyword. */
    template <class T> int getKeyword(std::string name, T &value); 

    /** Returns the number of the named column. */
    int         getColnum(char *name);
       
    /** Returns the number of rows in the current binary table. */
    long        getNumRows();          

    /** Write a value to a named column. */
    template <class T> int writeColumnValue(char *name, T value);

    /** Add new row to current table. */
    int         startNewRow();         

    /** Add new row to SUBINT table. */
    void        startNewSubintRow();   

    /** Write Primary Header Unit data. */
    void        writePHU(DataBlock &block);      

    /** Set up a FITS file */       
    int         initialise(DataBlock &block);

    /** Define and write the header keywords for the SUBINT table. */
    void        setupSubint(DataBlock &block);

    /** Write a row of SUBINT table data. */
    void        writeSubintRow(DataBlock &block);          

    /** Write information to HISTORY table. */
//     void        writeHistoryCommand(int argc, char * argv[]);
    void        writeHistoryCommand();
    void        writeHistoryTapeFile();

    /** Dealing with the command line arguments */
    void        setCommandLine(int argc, char *argv[]);

    /** Storing the name of the old-format data file */
    void        setOldFilename(std::string filename){oldFilename = filename;};


  private:
    fitsfile *fptr;           ///< The CFITSIO file variable
    std::string FITSname;     ///< The output FITS filename
    bool newFile;             ///< Is the file a new one?
    STATE filestate;          ///< OPEN or CLOSED?
    std::string UT_initial;   ///< The UT at the start of the FITS
			      ///   file data.
    std::string DATE_initial; ///< The UT date at the start of the
			      ///   data.
    int beam_number;          ///< What beam are we using? Necessary
			      ///   for FITS file name
    int blk_initial;          ///< The block number of the first block
			      ///   in the file.
    long numSamples;          ///< The number of samples stored in the
			      ///   FITS file.
    Subint subintTable;       ///< Subint class containing table
			      ///   header keywords.

    std::string oldFilename;  ///< The filename of the old-format data.
    std::string commandLine;  ///< The full command-line statement.

  };

}

#endif /*PSRFITS_HH_*/
