#include "pvo.hh"
#include "pvo-header.hh"
#include "pvo-block.hh"
#include "pvo-utils.hh"
#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <string>
#include <vector>

using std::string;
using namespace Header;
 
namespace Block
{

  DataBlock::DataBlock(SURVEY inputSurveyType)
  {
    /**
     * Initialise a block, by allocating the various arrays 
     *
     */
    this->headerString = string(" ",headerLength);
    Header::SurveySchema *temp = new Header::SurveySchema(inputSurveyType);
    this->survey = *temp;
    this->surveyType = inputSurveyType;
    this->data = new char[dataLength];
//     this->surveyName = survey;
    delete temp;
  }
  //--------------------------------------------------------------------------

  DataBlock::~DataBlock()
  {
    /** Deallocate survey and data first. */
    delete [] this->data;
  }
  //--------------------------------------------------------------------------

  bool DataBlock::readBlock(std::ifstream &file)
  {
    /**
     * Front end to the other two read functions.
     * \param file The file stream to read the block from.
     * \return False if readHeader() does not work. Otherwise true.
     */
    if(!this->readHeader(file)) return false;
    this->readData(file);
    return true;
  }
  //--------------------------------------------------------------------------

  bool DataBlock::readHeader(std::ifstream &file)
  {
    /**
     * Read in the full set of header information into the headerString string.
     * \param file The file stream to read the block from.
     * \return False if the file is at eof, else read header and return true.
     */
    char *hdr = new char[Block::headerLength];
    file.read( hdr, headerLength*sizeof(char) );
    bool returnval = true;;
    if(file.eof()) returnval = false;
    for(int c=0;c<headerLength;c++) this->headerString[c] = hdr[c];
    delete [] hdr;
    return returnval;
  }
  //--------------------------------------------------------------------------

  void DataBlock::readData(std::ifstream &file)
  {
    /**
     * Read in the full set of data into the data array (as type char).
     * \param file The file stream to read the block from.
     * \return No return value, as the file is assumed to be open.
     */
    file.read( this->data, dataLength*sizeof(char) );
  }
  //--------------------------------------------------------------------------

  void DataBlock::saveData(char *dataArray, int size)
  {
    /**
     * Saves an array of char values to the data array
     * \param dataArray The array of bytes to be saved.
     * \param size The size of dataArray. If less than
     * Block::dataLength, all elements will be written, but if more,
     * only Block::dataLength elements will be written
     * \return No return value.
     */

    int max = size;
    if(size>Block::dataLength) max = Block::dataLength;
    for(int i=0;i<max;i++) this->data[i] = dataArray[i];
    
  }
  //--------------------------------------------------------------------------

  std::string DataBlock::getEquinox()
  {
    /**
     * Returns the value of the equinox, assuming the coordinate type
     * is equatorial. If we are using Galactic or Ecliptic, a blank
     * string is returned.
     *\return A std::string containing the equinox value.
     */
    int type = this->getKeyword<int>(COORD_TYPE);
    std::string output;
    switch(type){
    case 2: output = "";               break;
    case 3: output = Block::eq_b1950;  break;
    case 4: output = "";               break;
    case 5:
    default: output = Block::eq_j2000; break;
    }
    return output;
  }
  //--------------------------------------------------------------------------

  std::string DataBlock::getCoordMode()
  {
    /**
     * Returns the value of the coord_md keyword as one of a number of
     * alternative strings. This keyword should be one of EQUAT, GAL,
     * ECLIP, etc. If the equinox is B1950 or J2000, return EQUAT.
     * \return A coord type: one of EQUAT, GAL, ECLIP...
     */
    int type = this->getKeyword<int>(COORD_TYPE);
    std::string output;
    switch(type){
    case 2: output = Block::eq_date;   break;
    case 3: output = Block::eq_equat;  break;
    case 4: output = Block::eq_gal;    break;
    case 5:
    default: output = Block::eq_equat; break;
    }
    return output;
  }
  //--------------------------------------------------------------------------

  bool DataBlock::telescopeOK()
  {
    /**
     * A way of determining whether the current block was taken at a time
     *  when the telescope was suitably positioned.
     *
     * The function tests the MOVE keyword (which is a coded integer),
     * and translates it according to the survey being conducted. There are
     * particular values of the keyword that are acceptable for each survey,
     * and if the keyword takes this value, everything is OK. Otherwise, the
     * telescope may have been moving, or not in survey mode.
     * \return True if everything is OK, false otherwise.
     */
    std::string move=this->getKeyword<std::string>(MOVE);

    if(this->surveyType == S70){
      if(move == "02") return false;
      else return true;
    }
    else if((this->surveyType == PM)||(this->surveyType == PH)){
      if(move[1]=='0') return true;
      else return false;
    }
    return false;
  }
  //--------------------------------------------------------------------------

  std::string DataBlock::getFITSname()
  {
    /**
     * Generates a string containing the name of the FITS file that the block
     * data should be written to. This depends on the survey, the date & time,
     * and (if the multibeam receiver was used) the beam number:
     * <ul> <li> S70_yymmdd_hhmmss.fits
     *      <li> PM_yymmdd_hhmmss_bb.fits </ul>
     * \return String containing the FITS filename.
     */
    std::string date = this->getKeyword<std::string>(DATE);
    std::string newdate;

    for(unsigned int i=0;i<date.size();i++)
      if(date[i]!='/') newdate += date[i];

    std::string time = this->getKeyword<std::string>(UT_STRT);
    time = removeLeadingBlanks(time);
    time = decToDMS( dmsToDec(time), 0, "");
    
    std::stringstream output;
    if(this->surveyType == S70) output << "S70_";
    else if(this->surveyType == PM) output << "PM_";
    else if(this->surveyType == PH) output << "PH_";
    else std::cerr << "Error <DataBlock::getFITSname> : unknown survey!\n";

    output << newdate << "_" << time;

    if((this->surveyType == PM)||(this->surveyType == PH))
      output << "_" 
	     << std::setw(2) << std::setfill('0') 
	     << this->getKeyword<int>(IBEAM);

    output << ".fits";

    return output.str();

  }
  //--------------------------------------------------------------------------
//   template <typename Type> 
//   Type DataBlock::getKeyword(int field, int element)
//   {
//     Type output = 
//       this->survey.readParameter<Type>(this->headerString,field,element);
//     return output;
//   }
//   template int DataBlock::getKeyword<int>(int field, int element);
//   template float DataBlock::getKeyword<float>(int field, int element);
//   template string DataBlock::getKeyword<string>(int field, int element);
  //--------------------------------------------------------------------------

  template <class T> 
  T DataBlock::getKeyword(int field, int elementNumber)
  {
    /**
     * Returns the value of the keyword. In the case of keywords that have more
     * than one value (some have 2 values, one for each band, and some have up 
     * to 8, describing individual comments etc), one can select which element
     * to return. This choice defaults to the first one if not given.
     *
     * The function is templated, so that you can elect to get the
     * numerical value or the string representation of the keyword.
     *
     * \param field The number representing the keyword field. This can be
     * represented by the enumerated list of values given by Header::FIELDS.
     * \param elementNumber Optional parameter indicating which element of a
     * multi-dimensional keyword to return. Defaults to 0, which is the first
     * element.
     */
    int dim = this->survey.getFieldDimension(field);
    std::stringstream ss;
    std::vector<T> output(dim);
    for(int i=0;i<dim;i++){
      int start = this->survey.getStartPosition(field)+i*dim;
      int length = this->survey.getFieldLength(field);
      ss.str(this->headerString.substr(start,length));
      ss >> output[i];      
    }
    return output[elementNumber];
  }
  template int DataBlock::getKeyword<int>(int field, int elementNumber);
  template float DataBlock::getKeyword<float>(int field, int elementNumber);
  template double DataBlock::getKeyword<double>(int field, int elementNumber);
  template std::string DataBlock::getKeyword<std::string>(int field, int elementNumber);
  
//   template<> int DataBlock::getKeyword<int>(int field, int element){
//     return this->survey.readParameter<int>(this->headerString,field,element);
//   }
//   template<> float DataBlock::getKeyword<float>(int field, int element){
//     return this->survey.readParameter<float>(this->headerString,field,element);
//   }
//   template<> double DataBlock::getKeyword<double>(int field, int element){
//     return this->survey.readParameter<double>(this->headerString,field,element);
//   }
//   template<> string DataBlock::getKeyword<string>(int field, int element){
//    return this->survey.readParameter<string>(this->headerString,field,element);
//   }
  //--------------------------------------------------------------------------
  template <class T> 
  std::vector<T> DataBlock::getFullKeyword(int field)
  {
    /**
     * Returns the full list of values of the keyword, in a vector of
     * the relevant (or requested) type. This function is useful to
     * obtain all values for the case of keywords that have more than
     * one value (some have 2 values, one for each band, and some have
     * up to 8, describing individual comments etc).
     *
     * The function is templated, so that you can elect to get the
     * numerical values or the string representations of individual
     * values of the keyword.
     *
     * \param field The number representing the keyword field. This can be
     * represented by the enumerated list of values given by Header::FIELDS.
     */
    int dim = this->survey.getFieldDimension(field);
    std::stringstream ss;
    std::vector<T> output(dim);
    for(int i=0;i<dim;i++){
      int start = this->survey.getStartPosition(field)+i*dim;
      int length = this->survey.getFieldLength(field);
      ss.str(this->headerString.substr(start,length));
      ss >> output[i];      
    }
    return output;
  }
  template<> std::vector<int> DataBlock::getFullKeyword<int>(int field){
    std::vector<int> output(this->survey.getFieldDimension(field));
    for(int i=0;i<this->survey.getFieldDimension(field);i++)
      output[i]=this->survey.readParameter<int>(this->headerString,field,i);
    return output;
  }
  template<> std::vector<float> DataBlock::getFullKeyword<float>(int field){
    std::vector<float> output(this->survey.getFieldDimension(field));
    for(int i=0;i<this->survey.getFieldDimension(field);i++)
      output[i]= this->survey.readParameter<float>(this->headerString,field,i);
    return output;
  }
  template<> std::vector<string> DataBlock::getFullKeyword<string>(int field){
    std::vector<string> output(this->survey.getFieldDimension(field));
    for(int i=0;i<this->survey.getFieldDimension(field);i++)
      output[i]=this->survey.readParameter<string>(this->headerString,field,i);
    return output;
  }

}

